port module App exposing (..)

import Browser
import Dict exposing (Dict)
import Html as H exposing (Html)
import Html.Attributes as HA
import Html.Events as HE
import Json.Decode as JD
import Json.Encode as JE
import LatLonDistance exposing (lat_lon_distance, LatLon)
import List.Extra as LE
import Tuple exposing (pair, first, second)

port receive_position : (JD.Value -> msg) -> Sub msg

port send_value : JD.Value -> Cmd msg

main = Browser.document
    { init = init
    , update = update
    , subscriptions = subscriptions
    , view = view
    }

type alias Marker =
  { pos : LatLon
  , icon : String
  , name : String
  , note : String
  }

type Selection
  = NoSelection
  | SelectedPosition LatLon
  | SelectedMarker Int
  | EditingMarker Int Marker

blank_marker =
  { pos = {lat = 0, lon = 0}
  , icon = "📌"
  , name = ""
  , note = ""
  }

type MapCentre
  = NoCentre
  | CurrentPositionCentre
  | PositionCentre LatLon

type alias Model = 
  { markers : List Marker
  , emoji : List Emoji
  , map_id : String
  , map_name : String
  , map_ids : List ({name: String, id: String})
  , current_position : LatLon
  , selection : Selection
  , new_marker : Marker
  , accuracy : Float
  , centre : MapCentre
  }

init_model = 
  { markers = []
  , emoji = []
  , map_id = ""
  , map_name = ""
  , map_ids = []
  , current_position = {lat = 55.04, lon = -1.46}
  , selection = NoSelection
  , new_marker = blank_marker
  , accuracy = 0
  , centre = CurrentPositionCentre
  }

type Msg
  = ReceivePosition JD.Value
  | MapClicked LatLon
  | MarkerClicked Int
  | AddMarker
  | UpdateNewMarker Marker
  | UpdateExistingMarker Int Marker
  | EditMarker Int Marker
  | RemoveMarker Int
  | ClickCurrentPosition
  | SetMapName String
  | Save

type alias Emoji =
  { emoji : String
  , description : String
  }

decode_emoji =
  JD.map2 Emoji
    (JD.field "emoji" JD.string)
    (JD.field "description" JD.string)

type alias Flags =
  { emoji : List Emoji
  , markers : List Marker
  , map_id : String
  , map_ids : List ({name: String, id: String})
  , map_name : String
  }

decode_flags =
  JD.map5 Flags
    (JD.field "emoji" (JD.list decode_emoji))
    (JD.at ["data", "markers"] (JD.list decode_marker))
    (JD.field "map_id" JD.string)
    (JD.field "map_ids" 
      (JD.list 
        (JD.map2 (\name id -> { name = name, id = id})
          (JD.field "name" JD.string)
          (JD.field "id" JD.string)
        )
      )
    )
    (JD.at ["data", "name"] JD.string)

init : (JD.Value) -> (Model, Cmd msg)
init vflags = 
  (case JD.decodeValue decode_flags vflags of
    Err err -> 
      let
        q = Debug.log "error" err
      in
        init_model
    Ok flags -> { init_model | emoji = flags.emoji, markers = flags.markers, map_id = flags.map_id, map_ids = flags.map_ids, map_name = flags.map_name }
  ) |> nocmd

nocmd m = (m, Cmd.none)

save : Model -> (Model, Cmd Msg)
save model = 
  (model
  , send_value
    <| JE.object
        [ ("type", JE.string "save")
        , ("data", JE.object
            [ ("markers", JE.list identity <| List.indexedMap (\i m -> encode_marker (String.fromInt i) m) model.markers)
            , ("name", JE.string model.map_name)
            ]
          )
        ]
  )

update msg model = 
  case msg of
    ReceivePosition v -> case JD.decodeValue decode_position v of
      Ok (pos, accuracy) -> { model | current_position = pos, accuracy = accuracy } |> nocmd
      _ -> model |> nocmd

    MapClicked pos -> case model.selection of
      NoSelection -> 
        let
          m = model.new_marker
        in
          { model | selection = SelectedPosition pos, new_marker = { m | pos = pos } } |> nocmd
      _ -> { model | selection = NoSelection } |> nocmd

    MarkerClicked i -> 
      let
        marker = LE.getAt i model.markers
        pos = marker |> Maybe.map .pos
        centre = case pos of
          Just p -> PositionCentre p
          Nothing -> model.centre
      in
        { model | selection = SelectedMarker i, centre = centre } |> nocmd

    AddMarker -> add_marker model |> save

    UpdateNewMarker marker -> { model | new_marker = marker } |> nocmd

    UpdateExistingMarker i marker -> { model | markers = LE.setAt i marker model.markers, selection = SelectedMarker i } |> save

    EditMarker i marker -> { model | selection = EditingMarker i marker } |> nocmd

    RemoveMarker i -> { model | markers = LE.removeAt i model.markers, selection = NoSelection } |> save

    ClickCurrentPosition -> { model | centre = CurrentPositionCentre, selection = NoSelection } |> nocmd

    SetMapName name -> { model | map_name = name } |> nocmd

    Save -> model |> save

add_marker model = case model.selection of
  SelectedPosition pos -> 
    let
      nm = model.new_marker
      marker = { nm | pos = pos }
    in
      { model | markers = marker::model.markers, selection = SelectedMarker 0, new_marker = blank_marker }

  _ -> model
  
-- For coordinates from the geolocation API
decode_latlon : JD.Decoder LatLon
decode_latlon = 
  JD.map2 LatLon
    (JD.field "latitude" JD.float)
    (JD.field "longitude" JD.float)

-- For coordinates from leafletjs
decode_latlng : JD.Decoder LatLon
decode_latlng = 
  JD.map2 LatLon
    (JD.field "lat" JD.float)
    (JD.field "lng" JD.float)

decode_position : JD.Decoder (LatLon, Float)
decode_position = JD.map2 pair decode_latlon (JD.field "accuracy" JD.float)

decode_map_click = 
     JD.at ["detail", "latlng"] decode_latlng
  |> JD.map MapClicked

encode_latlng : LatLon -> JE.Value
encode_latlng pos =
  JE.object
    [ ("lat", JE.float pos.lat)
    , ("lng", JE.float pos.lon)
    ]

decode_marker : JD.Decoder Marker
decode_marker =
  JD.map4 Marker
    (JD.field "pos" decode_latlng)
    (JD.field "icon" JD.string)
    (JD.field "name" JD.string)
    (JD.field "note" JD.string)

encode_marker : String -> Marker -> JD.Value
encode_marker id marker =
  JE.object
    [ ("id", JE.string id)
    , ("pos", encode_latlng marker.pos)
    , ("icon", JE.string marker.icon)
    , ("note", JE.string marker.note)
    , ("name", JE.string marker.name)
    ]

splitAt : String -> String -> (String, String)
splitAt sep str = 
  if str == "" then ("","") else
    if String.left (String.length sep) str == sep then
      ("", String.dropLeft (String.length sep) str)
    else
      let
        (a,b) = splitAt sep (String.dropLeft 1 str)
      in
        ((String.left 1 str)++a, b)
  

decode_marker_click =
     JD.at ["detail","id"] JD.string
  |> JD.andThen 
    (\id -> 
      let
        (l,r) = splitAt "-" id
      in
        case id of
          "current-position" -> JD.succeed ClickCurrentPosition
          _ -> case l of
            "marker" -> case String.toInt r of
              Just i -> JD.succeed (MarkerClicked i)
              Nothing -> JD.fail <| "Invalid int "++r
            _ -> JD.fail <| "Invalid marker type "++l
    )

subscriptions model =
  Sub.batch
    [ receive_position ReceivePosition
    ]

display_distance : Float -> String
display_distance d =
  let
    digits = max 1 (floor (logBase 10 d))
    magnitude = (10^digits)
  in
    if digits < 4 then (d/(toFloat magnitude) |> round |> (*) magnitude |> String.fromInt) ++ "m"
    else ((d/1000) |> round |> String.fromInt) ++ "km"

string_from_coords : LatLon -> String
string_from_coords c = (String.fromFloat c.lat) ++ "," ++ (String.fromFloat c.lon)

marker_link label i m = 
  H.a 
    [ HE.onClick (MarkerClicked i)
    , HA.href <| "#marker-"++(String.fromInt i)
    ]
    [ H.text label ]

space = H.text " "

view : Model -> Browser.Document Msg
view model = 
  let
    new_marker = model.new_marker 
    
    selected_marker : List (String, Marker)
    selected_marker = case model.selection of
      SelectedPosition pos -> [("",{ new_marker | pos = pos })]
      _ -> []

    existing_markers : List (String, Marker)
    existing_markers = List.indexedMap (\i m -> ("marker-"++(String.fromInt i), m)) model.markers

    current_position_marker = [("current-position", { blank_marker | icon = "😀", pos = model.current_position })]

    markers = 
         (current_position_marker ++ existing_markers ++ selected_marker)
      |> JE.list (\(id,m) -> encode_marker id m)
      |> JE.encode 0

    edit_marker_form marker update_msg submit extra_elements =
        H.form
          [ HE.onSubmit submit
          , HA.id "marker-form"
          , HA.class "marker-detail"
          ]
          ([ H.p [] [H.text <| Debug.toString marker.pos]
           , H.p
              []
              [ H.label [ HA.for "marker-name" ] [ H.text "Name" ]
              , H.input 
                [ HA.type_ "text"
                , HA.id "marker-name"
                , HA.value marker.name
                , HE.onInput (\s -> update_msg {marker | name = s })
                ]
                []
              , H.label [ HA.for "marker-icon" ] [ H.text "Icon" ]
              , H.input 
                [ HA.type_ "text"
                , HA.id "marker-icon"
                , HA.list "emoji"
                , HA.value marker.icon
                , HE.onInput (\s -> update_msg {marker | icon = s })
                ]
                []
              ]
            , H.p
              []
              [ H.label [ HA.for "marker-note" ] [ H.text "Note" ]
              , H.textarea 
                [ HA.id "marker-note"
                , HA.value marker.note
                , HE.onInput (\s -> update_msg {marker | note = s })
                ]
                []
              ]
            , H.datalist
              [ HA.id "emoji" ]
              (model.emoji |> List.map (\e -> H.option [HA.value e.emoji] [H.text <| e.emoji ++ " " ++ e.description]))
              
          ]++extra_elements)

    form = case model.selection of
      SelectedPosition pos -> 
        edit_marker_form 
          model.new_marker 
          UpdateNewMarker 
          AddMarker
          [ H.p [] [H.button [ HA.type_ "submit" ] [ H.text "Add a marker here" ]]
          ]
      
      SelectedMarker i -> case LE.getAt i model.markers of
        Just marker -> 
          H.div
            [ HA.class "marker-detail" ]
            [ H.h2 
              [] 
              [ H.button
                [ HA.type_ "button"
                , HE.onClick (EditMarker i marker)
                ]
                [ H.text "Edit" ]
              , space
              , H.span [ HA.class "icon" ] [H.text marker.icon ]
              , space
              , H.text marker.name
              ]
            , H.node "mark-down"
              []
              [ H.text marker.note ]
            ]
          
        Nothing -> H.div [] [H.text ""]
          
      EditingMarker i marker -> 
        edit_marker_form 
          marker 
          (EditMarker i) 
          (UpdateExistingMarker i marker)
          [ H.p 
            [] 
            [ H.button [ HA.type_ "submit" ] [ H.text "Update" ]
            , space
            , H.button [ HA.type_ "button", HE.onClick (RemoveMarker i) ] [ H.text "Delete" ]
            , space
            , marker_link "Cancel" i marker
            ]
          ]
      
      NoSelection ->
        H.div
          [ HA.id "menu"
          , HA.class "marker-detail"
          ]
          [ H.h1 [] [ H.text "Closest markers" ]
          , H.p
              []
              [ H.text <| "🗺️ "
              , H.form
                  [ HE.onSubmit Save
                  , HA.id "name-form"
                  ]
                  [ H.input
                    [ HA.type_ "text"
                    , HE.onInput SetMapName
                    , HA.value model.map_name
                    ]
                    []
                  ]
              ]
          , H.ul
            []
            (List.map (\(i,m) -> 
              H.li [] 
                [ marker_link (m.icon ++ " " ++ m.name) i m ]
              ) 
              (closest_markers |> List.take 3)
            )
          , H.p
            []
            [ H.a
                [ HA.href <| "?map=" ++ model.map_id ]
                [ H.text "🔗 Link to this map" ]
            ]
          , H.h3
              []
              [ H.text "Recently-used maps"]
          , H.ul
              []
              (List.map
                (\d -> H.li [] [H.a [HA.href <| "?map="++d.id] [H.text d.name, H.text " ", H.small [] [H.text d.id]]])
                model.map_ids
              )
          , H.p
              []
              [ H.a [HA.href "?"] [H.text "New map"]]
          ]

    closest_markers = 
      model.markers
      |> List.indexedMap pair
      |> List.sortBy (second >> .pos >> lat_lon_distance model.current_position)

    selection_str = case model.selection of
      NoSelection -> "none"
      SelectedPosition _ -> "position"
      SelectedMarker _ -> "viewing-marker"
      EditingMarker _ _ -> "editing-marker"

    (centre_mode, centre) = case model.centre of
      NoCentre -> (False, model.current_position)
      CurrentPositionCentre -> (True, model.current_position)
      PositionCentre pos -> (True, pos)
  in
    {
      title = "CLP's map",
      body = 
        [ H.main_
          [ HA.attribute "data-selection" selection_str
          ]
          [ form
          , H.node "leaflet-map" 
            [ HA.attribute "markers" <| markers
            , HA.attribute "lat" <| String.fromFloat centre.lat
            , HA.attribute "lon" <| String.fromFloat centre.lon
            , HA.attribute "centre" <| if centre_mode then "true" else "false"
            , HE.on "mapclick" decode_map_click
            , HE.on "markerclick" decode_marker_click
            ]
            []
          ]
        ]
    }